﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.ALPS {
  [Item("LayerCreator", "An operator which creates a new layer by cloning the oldest one and setting the variables accordingly.")]
  [StorableClass]
  public sealed class LayerCreator : SingleSuccessorOperator {
    private ILookupParameter<IntValue> OpenLayersParameter {
      get { return (ILookupParameter<IntValue>)Parameters["OpenLayers"]; }
    }

    [StorableConstructor]
    private LayerCreator(bool deserializing) : base(deserializing) { }

    private LayerCreator(LayerCreator original, Cloner cloner)
      : base(original, cloner) {
    }
    public LayerCreator()
      : base() {
      Parameters.Add(new LookupParameter<IntValue>("OpenLayers"));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new LayerCreator(this, cloner);
    }

    public override IOperation Apply() {
      var scopes = ExecutionContext.Scope.SubScopes;
      if (scopes.Count < 1)
        throw new ArgumentException("At least one sub-scope must exist.");

      var lastSubScope = scopes.Last();
      var clone = (IScope)lastSubScope.Clone(new Cloner());

      int number;
      if (int.TryParse(clone.Name, out number))
        clone.Name = (number + 1).ToString();

      scopes.Add(clone);

      // Set new layernumber
      clone.Variables["Layer"].Value = new IntValue(OpenLayersParameter.ActualValue.Value);

      // Decrement ages, because MainOperator is goint to increment it
      foreach (var solution in clone.SubScopes)
        ((DoubleValue)solution.Variables["Age"].Value).Value -= 1;

      // Reset existing values in the results to NaN to symbolize that no layer existed during that duration
      var results = (ResultCollection)clone.Variables["LayerResults"].Value;
      ResetResults(results);

      return base.Apply();
    }

    private void ResetResults(ResultCollection results) {
      var values = results.Select(r => r.Value);

      // Reset all values within results in results
      foreach (var resultsCollection in values.OfType<ResultCollection>())
        ResetResults(resultsCollection);

      // Reset values
      foreach (var dataTable in values.OfType<DataTable>()) {
        foreach (var row in dataTable.Rows)
          for (int i = 0; i < row.Values.Count; i++)
            row.Values[i] = double.NaN;
      }
    }

  }
}